'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;

  // Tracking variables build during the AST pass. We instantiate
  // these in the `Program` visitor in order to support running the
  // plugin in watch mode or on multiple files.
  var nanoutilss = void 0,
      specified = void 0,
      selectedMethods = void 0;

  // Import a nanoutils method and return the computed import identifier
  function importMethod(methodName, file, cjs) {
    if (!selectedMethods[methodName]) {
      var path = (0, _modules2.default)(methodName, cjs);
      selectedMethods[methodName] = file.addImport(path, 'default');
    }
    return t.clone(selectedMethods[methodName]);
  }

  function matchesNanoutils(path, name) {
    return nanoutilss[name] && (hasBindingOfType(path.scope, name, 'ImportDefaultSpecifier') || hasBindingOfType(path.scope, name, 'ImportNamespaceSpecifier'));
  }

  function matchesNanoutilsMethod(path, name) {
    return specified[name] && hasBindingOfType(path.scope, name, 'ImportSpecifier');
  }

  function hasBindingOfType(scope, name, type) {
    return scope.hasBinding(name) && scope.getBinding(name).path.type === type;
  }

  return {
    visitor: {
      Program: {
        enter: function enter() {
          // Track the variables used to import nanoutils
          nanoutilss = Object.create(null);
          specified = Object.create(null);
          // Track the methods that have already been used to prevent dupe imports
          selectedMethods = Object.create(null);
        }
      },
      ImportDeclaration: function ImportDeclaration(path) {
        var node = path.node;

        if (node.source.value === 'nanoutils') {
          node.specifiers.forEach(function (spec) {
            if (t.isImportSpecifier(spec)) {
              specified[spec.local.name] = spec.imported.name;
            } else {
              nanoutilss[spec.local.name] = true;
            }
          });
          path.remove();
        }
      },
      ExportNamedDeclaration: function ExportNamedDeclaration(path, state) {
        var node = path.node,
            hub = path.hub;

        if (node.source && node.source.value === 'nanoutils') {
          var specifiers = node.specifiers.map(function (spec) {
            var importIdentifier = importMethod(spec.exported.name, hub.file, state.opts.cjs);
            var exportIdentifier = t.identifier(spec.local.name);
            return t.exportSpecifier(importIdentifier, exportIdentifier);
          });
          node.specifiers = specifiers;
          node.source = null;
        }
      },
      ExportAllDeclaration: function ExportAllDeclaration(path) {
        var node = path.node;

        if (node.source && node.source.value === 'nanoutils') {
          throw new Error('`export * from "nanoutils"` defeats the purpose of babel-plugin-nanoutils');
        }
      },
      CallExpression: function CallExpression(path, state) {
        var node = path.node,
            hub = path.hub;
        var name = node.callee.name;

        if (!t.isIdentifier(node.callee)) return;
        if (matchesNanoutilsMethod(path, name)) {
          node.callee = importMethod(specified[name], hub.file, state.opts.cjs);
        }
        if (node.arguments) {
          node.arguments = node.arguments.map(function (arg) {
            var name = arg.name;

            return matchesNanoutilsMethod(path, name) ? importMethod(specified[name], hub.file, state.opts.cjs) : arg;
          });
        }
      },
      MemberExpression: function MemberExpression(path, state) {
        var node = path.node;

        var objectName = node.object.name;
        if (!matchesNanoutils(path, objectName)) return;
        // R.foo() -> foo()
        var newNode = importMethod(node.property.name, path.hub.file, state.opts.cjs);
        path.replaceWith({ type: newNode.type, name: newNode.name });
      },
      Property: function Property(path, state) {
        var node = path.node,
            hub = path.hub;

        if (t.isIdentifier(node.key) && node.computed && matchesNanoutilsMethod(path, node.key.name)) {
          node.key = importMethod(specified[node.key.name], hub.file, state.opts.cjs);
        }
        if (t.isIdentifier(node.value) && matchesNanoutilsMethod(path, node.value.name)) {
          node.value = importMethod(specified[node.value.name], hub.file, state.opts.cjs);
        }
      },
      Identifier: function Identifier(path, state) {
        var node = path.node,
            hub = path.hub,
            parent = path.parent,
            scope = path.scope;
        var name = node.name;

        if (matchesNanoutilsMethod(path, name) && !isSpecialTypes(t, parent)) {
          var newNode = importMethod(specified[name], hub.file, state.opts.cjs);
          path.replaceWith({ type: newNode.type, name: newNode.name });
        } else if (matchesNanoutils(path, name)) {
          // #19, nullify direct references to the nanoutils import (for apply/spread/etc)
          var replacementNode = t.nullLiteral();
          path.replaceWith(replacementNode);
        }
      }
    }
  };
};

var _modules = require('./modules');

var _modules2 = _interopRequireDefault(_modules);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var SPECIAL_TYPES = ['isMemberExpression', 'isProperty'];

function isSpecialTypes(t, node) {
  return SPECIAL_TYPES.filter(function (type) {
    return t[type](node);
  }).length > 0;
}